#!/usr/bin/env python
# -*- coding: utf-8 -*-
###############################################################################
# (c) Copyright 2000-2019 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
# =============================================================================
## @file Models.py
#
#  Module with some useful utilities for simple functions and fit models.
#
#  @author Vanya BELYAEV Ivan.Belyaev@cern.ch
#  @date   2011-12-01
# =============================================================================
"""Module with some useful fit-models"""
# =============================================================================
__author__  = "Vanya BELYAEV Ivan.Belyaev@cern.ch"
__date__    = "2011-12-01"
__all__     = ()
# =============================================================================
import  ROOT 
from    Ostap.Core import funID, cpp 
# =============================================================================
# logging 
# =============================================================================
from   Ostap.Logger import getLogger 
if '__main__' ==  __name__ : logger = getLogger ( 'Ostap.Models' )
else                       : logger = getLogger ( __name__       )
# =============================================================================
# helper adapter for 1D-functions 
class _WO1_ (object)  :
    "Helper adapter for 1D-functions"
    def __init__ ( self , o              ) :        self._o   =  o 
    def __call__ ( self , x , pars  = [] ) : return self._o ( x [0]        )
# helper adapter for 2D-functions 
class _WO2_ (object)  :
    "Helper adapter for 2D-functions"
    def __init__ ( self , o              ) :        self._o   =  o 
    def __call__ ( self , x , pars  = [] ) : return self._o ( x [0] , x[1] )
# helper adapter for 3D-functions 
class _WO3_ (object)  :
    "Helper adapter for 3D-functions"
    def __init__ ( self , o              ) :        self._o   =  o 
    def __call__ ( self , x , pars  = [] ) : return self._o ( x [0] , x[1] , x[2] )

# =============================================================================
pos_infinity = float('+inf')
neg_infinity = float('-inf')
# =============================================================================
## convert the model into TF1
def _tf1_ ( self                 ,
            xmin  = neg_infinity ,
            xmax  = pos_infinity ,
            npars = 0            , *args ) :
    """Convert ``the function'' to ROOT.TF1
    
    >>> obj = ...
    
    >>> fun = obj.tf1 ( 3.0 , 3.2 )
    
    >>> fun.Draw() 
    """
    #
    if not hasattr ( self , '_wo1' ) : self._wo1 = _WO1_ ( self )
    if not self._wo1                 : self._wo1 = _WO1_ ( self )
    #
    if hasattr ( self , 'xmin'  ) : xmin  = max ( xmin  , self.xmin () )
    if hasattr ( self , 'xmax'  ) : xmax  = min ( xmax  , self.xmax () )
    if hasattr ( self , 'npars' ) : npars = max ( npars , self.npars() )
    #
    _wo = self._wo1 
    fun = ROOT.TF1 ( funID()  , _wo , xmin , xmax , npars, *args )
    fun.SetNpx ( 500 ) 
    #
    return fun 

# =============================================================================
## convert the model into TF2
def _tf2_ ( self ,
            xmin  = neg_infinity ,
            xmax  = pos_infinity ,
            ymin  = neg_infinity ,
            ymax  = pos_infinity ,
            npars = 0            , *args ) :
    """Convert ``the function'' to ROOT.TF2
    
    >>> obj = ...
    
    >>> fun = obj.tf2 ( 3.0 , 3.2 , 3.0 , 3.2 )
    
    >>> fun.Draw() 
    """
    ##
    if not hasattr ( self , '_wo2' ) : self._wo2 = _WO2_ ( self )
    if not self._wo2                 : self._wo2 = _WO2_ ( self )
    ## 
    if hasattr ( self , 'xmin'  ) : xmin  = max ( xmin  , self.xmin () )
    if hasattr ( self , 'xmax'  ) : xmax  = min ( xmax  , self.xmax () )
    if hasattr ( self , 'ymin'  ) : ymin  = max ( ymin  , self.ymin () )
    if hasattr ( self , 'ymax'  ) : ymax  = min ( ymax  , self.ymax () )
    if hasattr ( self , 'npars' ) : npars = max ( npars , self.npars() )
    #
    _wo = self._wo2
    fun = ROOT.TF2 ( funID ()  , _wo , xmin , xmax , ymin , ymax , npars , *args )
    fun.SetNpx ( 100 ) 
    fun.SetNpy ( 100 ) 
    #
    return fun 

# =============================================================================
## convert the model into TF3
def _tf3_ ( self ,
            xmin  = neg_infinity ,
            xmax  = pos_infinity ,
            ymin  = neg_infinity ,
            ymax  = pos_infinity ,
            zmin  = neg_infinity ,
            zmax  = pos_infinity ,
            npars = 0            , *args ) :
    """Convert ``the function'' to TF3  
    >>> obj = ...
    
    >>> fun = obj.tf3 ( 3.0 , 3.2 , 3.0 , 3.2 , 1 , 2 )
    
    >>> fun.Draw() 
    """
    ##
    if not hasattr ( self , '_wo3' ) : self._wo3 = _WO3_ ( self )
    if not self._wo3                 : self._wo3 = _WO3_ ( self )
    ## 
    if hasattr ( self , 'xmin'  ) : xmin  = max ( xmin  , self.xmin () )
    if hasattr ( self , 'xmax'  ) : xmax  = min ( xmax  , self.xmax () )
    if hasattr ( self , 'ymin'  ) : ymin  = max ( ymin  , self.ymin () )
    if hasattr ( self , 'ymax'  ) : ymax  = min ( ymax  , self.ymax () )
    if hasattr ( self , 'zmin'  ) : zmin  = max ( zmin  , self.zmin () )
    if hasattr ( self , 'zmax'  ) : zmax  = min ( zmax  , self.zmax () )
    if hasattr ( self , 'npars' ) : npars = max ( npars , self.npars() )
    #
    _wo = self._wo3
    fun = ROOT.TF3 ( funID ()  , _wo ,
                     xmin  , xmax  ,
                     ymin  , ymax  ,
                     zmin  , zmax  ,
                     npars , *args )
    fun.SetNpx ( 50 ) 
    fun.SetNpy ( 50 ) 
    fun.SetNpz ( 50 ) 
    #
    return fun 

# =============================================================================
## draw the function 
def _f1_draw_ ( self , *opts ) :
    """Drawing the function object through conversion to ROOT.TF1
    
    >>> fun = ...
    >>> fun.draw()
    
    """
    if not hasattr ( self , '_tf1'  ) :

        self._tf1 = _tf1_ ( self )
        if type(self) in ( Gaudi.Math.Positive          ,
                           Gaudi.Math.PositiveEven      , 
                           Gaudi.Math.Monothonic        , 
                           Gaudi.Math.Convex            , 
                           Gaudi.Math.ConvexOnly        , 
                           Gaudi.Math.PositiveSpline    , 
                           Gaudi.Math.MonothonicSpline  , 
                           Gaudi.Math.ConvexSpline      ,
                           Gaudi.Math.ConvexOnlySpline  ,
                           Gaudi.Math.ExpoPositive      ,
                           Gaudi.Math.TwoExpoPositive   ) :                                
            self._tf1.SetMinimum(0)
            
    return self._tf1.Draw ( *opts )
        
# =============================================================================
## get the regular complex value for amplitude 
def _amp_ ( self , x ) :
    """ Get the complex value for amplitude
    >>> fun
    >>> a = fun.amp ( x )    
    """
    v = self.amplitude ( x )
    #
    return complex( v.real () , v.imag () ) 

Gaudi = cpp.Gaudi 

Gaudi.Math.LASS        . amp = _amp_
Gaudi.Math.LASS23L     . amp = _amp_
Gaudi.Math.Bugg23L     . amp = _amp_
Gaudi.Math.Flatte      . amp = _amp_
Gaudi.Math.Flatte2     . amp = _amp_
Gaudi.Math.Flatte23L   . amp = _amp_
Gaudi.Math.BreitWigner . amp = _amp_
Gaudi.Math.Swanson     . amp = _amp_

# =============================================================================
## make 1D-integration using SciPy
#  @see http://www.scipy.org/
#  @author Vanya BELYAEV Ivan.Belyaev@itep.ru
#  @date 2014-12-01
def sp_integrate_1D ( func , xmin , xmax , *args , **kwargs ) :
    """Make 1D-integration using SciPy
    
    >>> func = ...
    >>> print func.sp_integrate ( -10 , 10 )
    
    """    
    from scipy import integrate
    ##
    result = integrate.quad ( func , xmin , xmax , *args , **kwargs )
    return result[0]

# =============================================================================
## make 2D-integration using SciPy
#  @see http://www.scipy.org/
#  @author Vanya BELYAEV Ivan.Belyaev@itep.ru
#  @date 2014-12-01
def sp_integrate_2D ( func  ,
                      xmin  , xmax ,
                      ymin  , ymax , *args , **kwargs ) :
    """Make 2D-integration using SciPy

    >>> func = ...  ## func ( x , y )
    ##                            xmin , xmax , ymin , ymax 
    >>> print func.sp_integrate ( -10  , 10   , -20  , 20   ) 
    
    Note different naming with respect to SciPy:
    - SciPy first integrates over 2nd variable
    """
    from scipy import integrate
    ##
    result = integrate.dblquad ( func ,
                                 ymin ,
                                 ymax ,
                                 lambda x : xmin ,
                                 lambda x : xmax , 
                                 *args , **kwargs )
    return result[0]



# =============================================================================
## make 3D-integration using SciPy
#  @see http://www.scipy.org/
#  @author Vanya BELYAEV Ivan.Belyaev@itep.ru
#  @date 2017-11-22
def sp_integrate_3D ( func  ,
                      xmin  , xmax ,
                      ymin  , ymax ,
                      zmin  , zmax , *args , **kwargs ) :
    """Make 2D-integration using SciPy

    >>> func = ...  ## func ( x , y , z )
    ##                            xmin , xmax , ymin , ymax  zmin , zmax 
    >>> print func.sp_integrate ( -10  , 10   , -20  , 20   , -1  , 3 ) 
    
    Note different naming with respect to SciPy:
    - SciPy first integrates over 2nd variable
    """
    from scipy import integrate
    ##
    result = integrate.tplquad ( func ,
                                 zmin ,
                                 zmax ,
                                 lambda   z : ymin ,
                                 lambda   z : ymax ,
                                 lambda y,z : xmin ,
                                 lambda y,z : xmax ,
                                 *args , **kwargs )
    return result[0]



# =============================================================================
## make 1D-integration using SciPy
#  @see http://www.scipy.org/
#  @author Vanya BELYAEV Ivan.Belyaev@itep.ru
#  @date 2014-12-01
def sp_integrate_1D_ ( pdf , xmin , xmax , *args , **kwargs ) :
    """Make 1D integration over the PDF using SciPy
    """
    if hasattr ( pdf , 'setPars' ) : pdf.setPars() 
    func = pdf.function()
    return func.sp_integrate_1D ( xmin , xmax , *args , **kwargs ) 

# =============================================================================
## make 2D-integration using SciPy
#  @see http://www.scipy.org/
#  @author Vanya BELYAEV Ivan.Belyaev@itep.ru
#  @date 2014-12-01
def sp_integrate_2D_ ( pdf   ,
                      xmin  , xmax ,
                      ymin  , ymax , *args , **kwargs ) :
    """ Make 2D integration over the PDF using SciPy
    """
    if hasattr ( pdf , 'setPars' ) : pdf.setPars() 
    func = pdf.function()
    return func.sp_integrate_2D ( xmin , xmax , ymin , ymax , *args , **kwargs ) 

# =============================================================================
## make 2D-integration using SciPy
#  @see http://www.scipy.org/
#  @author Vanya BELYAEV Ivan.Belyaev@itep.ru
#  @date 2014-12-01
def sp_integrate_3D_  ( pdf   ,
                        xmin  , xmax ,
                        ymin  , ymax ,
                        zmin  , zmax ,
                        *args , **kwargs ) :
    """ Make 3D integration over the PDF using SciPy
    """
    if hasattr ( pdf , 'setPars' ) : pdf.setPars() 
    func = pdf.function()
    return func.sp_integrate_3D ( xmin , xmax , ymin , ymax , zmin , zmax , *args , **kwargs ) 


from LHCbMath.deriv import moment   as sp_moment
from LHCbMath.deriv import mean     as sp_mean
from LHCbMath.deriv import variance as sp_variance
from LHCbMath.deriv import rms      as sp_rms 
from LHCbMath.deriv import median   as sp_median
from LHCbMath.deriv import quantile as sp_quantile
from LHCbMath.deriv import mode     as sp_mode 


# =============================================================================
## helper function to delegate some methods/attributes to TF1
#  @code
#  f = ...
#  f.SetLineColor(4) ## delegate to TF1
#  f.SetLineWidth(2) ## delegate to TF1
#  @endcode 
def _f1_getattr_ ( self , attr ) :
    """Delegate some methods/attributes to TF1
    >>> f = ...
    >>> f.SetLineColor(4) ## delegate to TF1
    >>> f.SetLineWidth(2) ## delegate to TF1
    """
    if  hasattr ( ROOT.TF1 , attr ) :
        if not hasattr  ( self      , '_tf1' ) : self._tf1 = self.tf1 ( ) 
        return getattr  ( self._tf1 , attr   )
    
    raise AttributeError

# =============================================================================
## helper function to delegate some methods/attributes to TF2
#  @code
#  f = ...
#  f.SetLineColor(4) ## delegate to TF2
#  f.SetLineWidth(2) ## delegate to TF2
#  @endcode 
def _f2_getattr_ ( self , attr ) :
    """Delegate some methods/attributes to TF2
    >>> f = ...
    >>> f.SetLineColor(4) ## delegate to TF2
    >>> f.SetLineWidth(2) ## delegate to TF2
    """
    if  hasattr ( ROOT.TF2 , attr ) :
        if not hasattr  ( self      , '_tf2' ) : self._tf2 = self.tf2 ( ) 
        return getattr  ( self._tf2 , attr   )
    
    raise AttributeError

# =============================================================================
## helper function to delegate some methods/attributes to TF3
#  @code
#  f = ...
#  f.SetLineColor(4) ## delegate to TF3
#  f.SetLineWidth(2) ## delegate to TF3
#  @endcode 
def _f3_getattr_ ( self , attr ) :
    """Delegate some methods/attributes to TF2
    >>> f = ...
    >>> f.SetLineColor(4) ## delegate to TF2
    >>> f.SetLineWidth(2) ## delegate to TF2
    """
    if  hasattr ( ROOT.TF3 , attr ) :
        if not hasattr  ( self      , '_tf3' ) : self._tf3 = self.tf3 ( ) 
        return getattr  ( self._tf3 , attr   )
    
    raise AttributeError


# =============================================================================
## get min/max values for bernstein polynomials
#  @code
#  p = ...
#  mn,mx = p.minmax()
#  @endcode
#  The values are guaranteed that
#  mn <= p(x) <= mx for all   xmin <= x <= xmax 
def _b_minmax_ ( bp ) :
    """Get min/max values for bernstein polynomials
    
    >>> p = ...
    >>> mn,mx = p.minmax()

    The values are such that: mn <= p(x) <= mx for all x_min<=x<x_max
    """
    b    = bp.bernstein() 
    pars = b .pars()
    mn   = min ( pars )
    mx   = max ( pars )
    return mn , mx 

# ==============================================================================
## get the maximal value for bernstein polynomial:
#  @code
#  p  = ...
#  mx = p.max()
#  @endcode
#  The values are guaranteed that
#  p(x) <= mx for all   xmin <= x <= xmax 
def _b_max_ ( bp ) :
    """Get max values for bernstein polynomials
    
    >>> p = ...
    >>> mx = p.max()

    The value is such that: p(x) <= mx  for all x_min<=x<x_max
    """
    b    = bp.bernstein() 
    pars = b .pars()
    return max ( pars )

# ==============================================================================
## get the minimal value for bernstein polynomial:
#  @code
#  p  = ...
#  mn = p.min()
#  @endcode
#  The values are guaranteed that
#  mn <= p(x) for all   xmin <= x <= xmax 
def _b_min_ ( bp ) :
    """Get min values for bernstein polynomials
    
    >>> p  = ...
    >>> mn = p.min()

    The  value is such that: mn <= p(x)  for all x_min<=x<x_max
    """
    b    = bp.bernstein() 
    pars = b .pars()
    return min ( pars )

for t in (  Gaudi.Math.Bernstein     ,
            Gaudi.Math.BernsteinEven ) :
    if not hasattr ( t , 'min'    ) : t.min    = _b_min_
    if not hasattr ( t , 'max'    ) : t.max    = _b_max_ 
    if not hasattr ( t , 'minmax' ) : t.minmax = _b_minmax_

# =============================================================================

# =============================================================================
## get min/max values for derived bernstein polynomials
#  @code
#  p = ...
#  mn,mx = p.minmax()
#  @endcode
#  The values are guaranteed that
#  mn <= p(x) <= mx for all   xmin <= x <= xmax 
def _p_minmax_ ( p ) :
    """Get min/max values for derived bernstein polynomials
    
    >>> p = ...
    >>> mn,mx = p.minmax()

    The values are such that: mn <= p(x) <= mx  for all x_min<=x<x_max
    """
    b    = p .bernstein() 
    pars = b .pars()
    mn   = min ( pars )
    mx   = max ( pars )
    return  max ( mn , 0 ) , mx 


# ==============================================================================
## get the minimal value for derived bernstein polynomial:
#  @code
#  p  = ...
#  mn = p.min()
#  @endcode
#  The values are guaranteed that
#  mn <= p(x) for all   xmin <= x <= xmax 
def _p_min_ ( bp ) :
    """Get min values for derived bernstein polynomials
    
    >>> p  = ...
    >>> mn = p.min()

    The  value is such that: mn <= p(x)  for all x_min<=x<x_max
    """
    b    = bp.bernstein() 
    pars = b .pars()
    return max ( min ( pars ) , 0 )

for t in (  Gaudi.Math.Positive      ,
            Gaudi.Math.Monothonic    ,
            Gaudi.Math.Convex        ,
            Gaudi.Math.ConvexOnly    ) :
    if not hasattr ( t , 'min'    ) : t.min    = _p_min_
    if not hasattr ( t , 'max'    ) : t.max    = _b_max_   ## ATTENTION: "b" is here
    if not hasattr ( t , 'minmax' ) : t.minmax = _p_minmax_

# =============================================================================
## try to get max-value using exisint mode function
#  @code
#  f = ...
#  mx = f.max()
#  @endcode
def _f_max_mode_ ( f ) :
    """Get the max-value (using exising mode function)
    >>> f = ...
    >>> mx = f.max()    
    """
    return f ( f.mode() ) 
    
# =============================================================================
## decorate 1D-models/functions 
# =============================================================================
Gaudi = cpp.Gaudi 
for model in ( Gaudi.Math.Chebyshev              ,
               Gaudi.Math.ChebyshevU             ,
               Gaudi.Math.Legendre               ,
               Gaudi.Math.Hermite                ,
               Gaudi.Math.Bernstein              ,
               Gaudi.Math.BernsteinEven          ,
               Gaudi.Math.ChebyshevSum           ,
               Gaudi.Math.LegendreSum            ,
               Gaudi.Math.HermiteSum             ,
               Gaudi.Math.FourierSum             ,
               Gaudi.Math.CosineSum              ,
               Gaudi.Math.Polynomial             ,               
               Gaudi.Math.Positive               ,
               Gaudi.Math.PositiveEven           ,
               Gaudi.Math.Monothonic             ,
               Gaudi.Math.Convex                 ,
               Gaudi.Math.ConvexOnly             ,
               Gaudi.Math.BifurcatedGauss        ,
               Gaudi.Math.Bukin                  ,
               Gaudi.Math.Novosibirsk            ,
               Gaudi.Math.CrystalBall            ,
               Gaudi.Math.Needham                ,
               Gaudi.Math.CrystalBallDoubleSided ,
               Gaudi.Math.GramCharlierA          ,
               Gaudi.Math.PhaseSpace2            ,
               Gaudi.Math.PhaseSpaceLeft         ,
               Gaudi.Math.PhaseSpaceRight        ,
               Gaudi.Math.PhaseSpaceNL           ,
               Gaudi.Math.PhaseSpace23L          ,
               Gaudi.Math.BreitWigner            ,
               Gaudi.Math.Rho0                   ,
               Gaudi.Math.Kstar0                 ,
               Gaudi.Math.Phi0                   ,
               Gaudi.Math.Rho0FromEtaPrime       ,
               Gaudi.Math.Flatte                 ,
               Gaudi.Math.Flatte2                ,
               Gaudi.Math.LASS                   ,
               Gaudi.Math.LASS23L                ,
               Gaudi.Math.Bugg23L                ,
               Gaudi.Math.BW23L                  ,
               Gaudi.Math.Flatte23L              ,
               Gaudi.Math.Gounaris23L            ,
               Gaudi.Math.StudentT               ,
               Gaudi.Math.BifurcatedStudentT     ,
               Gaudi.Math.Voigt                  ,
               Gaudi.Math.PseudoVoigt            ,
               Gaudi.Math.Logistic               ,
               Gaudi.Math.Spline                 ,
               Gaudi.Math.Linear                 ,
               Gaudi.Math.Parabola               ,
               #
               Gaudi.Math.GenGaussV1             ,
               Gaudi.Math.GenGaussV2             ,
               Gaudi.Math.SkewGauss              ,
               Gaudi.Math.GammaDist              ,
               Gaudi.Math.GenGammaDist           ,
               Gaudi.Math.Amoroso                ,
               Gaudi.Math.LogGammaDist           ,
               Gaudi.Math.Log10GammaDist         ,
               Gaudi.Math.LogGamma               ,
               Gaudi.Math.BetaPrime              ,
               Gaudi.Math.Landau                 ,
               Gaudi.Math.JohnsonSU              ,
               Gaudi.Math.Atlas                  ,
               Gaudi.Math.Sech                   ,
               Gaudi.Math.Swanson                ,
               Gaudi.Math.Argus                  ,
               Gaudi.Math.Tsallis                ,
               Gaudi.Math.QGSM                   ,
               Gaudi.Math.TwoExpos               ,
               Gaudi.Math.Sigmoid                ,
               #
               Gaudi.Math.BSpline                , 
               Gaudi.Math.PositiveSpline         ,
               Gaudi.Math.MonothonicSpline       ,
               Gaudi.Math.ConvexSpline           ,
               Gaudi.Math.ConvexOnlySpline       ,
               #
               Gaudi.Math.BernsteinDualBasis     ,

               ) :
    model.tf1             = _tf1_ 
    model.tf              = _tf1_ 
    model.sp_integrate    = sp_integrate_1D
    model.sp_integrate_1D = sp_integrate_1D
    model.__getattr__     = _f1_getattr_
    
    if not hasattr ( model , 'max' ) : 
        if hasattr ( model , 'mode' ) :
            model.max = _f_max_mode_ 

    if not hasattr ( model , 'mean'     ) : model.mean     = sp_mean 
    if not hasattr ( model , 'variance' ) : model.variance = sp_variance 
    if not hasattr ( model , 'rms'      ) : model.rms      = sp_rms  
    if not hasattr ( model , 'median'   ) : model.median   = sp_median
    if not hasattr ( model , 'mode'     ) : model.mode     = sp_mode 
    if not hasattr ( model , 'moment'   ) : model.moment   = sp_moment
    if not hasattr ( model , 'quantile' ) : model.quantile = sp_quantile 
    
        
## add some drawing method for some shapes 
for model in ( Gaudi.Math.Bernstein         ,
               Gaudi.Math.BernsteinEven     , 
               Gaudi.Math.Positive          ,
               Gaudi.Math.PositiveEven      ,
               Gaudi.Math.Monothonic        ,
               Gaudi.Math.Convex            ,
               Gaudi.Math.ConvexOnly        ,
               Gaudi.Math.ChebyshevSum      ,               
               Gaudi.Math.LegendreSum       ,               
               Gaudi.Math.HermiteSum        ,               
               Gaudi.Math.FourierSum        ,               
               Gaudi.Math.CosineSum         ,               
               Gaudi.Math.Polynomial        ,               
               Gaudi.Math.ExpoPositive      , 
               Gaudi.Math.TwoExpoPositive   , 
               #
               Gaudi.Math.BSpline           ,
               Gaudi.Math.MonothonicSpline  ,
               Gaudi.Math.PositiveSpline    , 
               Gaudi.Math.ConvexSpline      , 
               Gaudi.Math.ConvexOnlySpline  ) : 
    
    model.draw = _f1_draw_
    model.Draw = _f1_draw_

# =============================================================================
def _f_print_ ( self , typ = '' ) :
    if not typ : typ = str(type(self))
    return '%s(%s,%s,%s)' % ( typ ,  self.pars() , self.xmin() , self.xmax() )

Gaudi.Math.LegendreSum   .__str__  = lambda s : _f_print_ ( s , 'LegendreSum'   )
Gaudi.Math.ChebyshevSum  .__str__  = lambda s : _f_print_ ( s , 'ChebyshevSum'  )
Gaudi.Math.Polynomial    .__str__  = lambda s : _f_print_ ( s , 'Polynomial'    )
Gaudi.Math.Bernstein     .__str__  = lambda s : _f_print_ ( s , 'Bernstein'     )
Gaudi.Math.BernsteinEven .__str__  = lambda s : _f_print_ ( s , 'BernsteinEven' )
Gaudi.Math.Positive      .__str__  = lambda s : _f_print_ ( s , 'Positive'      )
Gaudi.Math.PositiveEven  .__str__  = lambda s : _f_print_ ( s , 'PositiveEven'  )
Gaudi.Math.FourierSum    .__str__  = lambda s : _f_print_ ( s , 'FourierSum'    )
Gaudi.Math.CosineSum     .__str__  = lambda s : _f_print_ ( s , 'CosineSum'     )

Gaudi.Math.LegendreSum   .__repr__ = lambda s : _f_print_ ( s , 'LegendreSum'   )
Gaudi.Math.ChebyshevSum  .__repr__ = lambda s : _f_print_ ( s , 'ChebyshevSum'  )
Gaudi.Math.HermiteSum    .__repr__ = lambda s : _f_print_ ( s , 'HermiteSum'    )
Gaudi.Math.Polynomial    .__repr__ = lambda s : _f_print_ ( s , 'Polynomial'    )
Gaudi.Math.Bernstein     .__repr__ = lambda s : _f_print_ ( s , 'Bernstein'     )
Gaudi.Math.BernsteinEven .__repr__ = lambda s : _f_print_ ( s , 'BernsteinEven' )
Gaudi.Math.Positive      .__repr__ = lambda s : _f_print_ ( s , 'Positive'      )
Gaudi.Math.PositiveEven  .__repr__ = lambda s : _f_print_ ( s , 'PositiveEven'  )
Gaudi.Math.Convex        .__repr__ = lambda s : _f_print_ ( s , 'Convex'        ) 
Gaudi.Math.ConvexOnly    .__repr__ = lambda s : _f_print_ ( s , 'ConvexOnly'    )
Gaudi.Math.Monothonic    .__repr__ = lambda s : _f_print_ ( s , 'Monothonic'    )
Gaudi.Math.FourierSum    .__repr__ = lambda s : _f_print_ ( s , 'FourierSum'    )
Gaudi.Math.CosineSum     .__repr__ = lambda s : _f_print_ ( s , 'CosineSum'     )


# =============================================================================
## decorate 2D-models/functions 
# =============================================================================
Gaudi = cpp.Gaudi 
for model in ( Gaudi.Math.Spline2D       ,
               Gaudi.Math.Spline2DSym    , 
               Gaudi.Math.Bernstein2D    ,
               Gaudi.Math.Positive2D     ,
               Gaudi.Math.Bernstein2DSym ,
               Gaudi.Math.Positive2DSym  ,
               Gaudi.Math.PS2DPol        ,
               Gaudi.Math.PS2DPolSym     ,
               Gaudi.Math.ExpoPS2DPol    ,
               Gaudi.Math.Expo2DPol      ,
               Gaudi.Math.Expo2DPolSym   ) :
    
    model . tf2 = _tf2_ 
    model . tf  = _tf2_ 
    model.__getattr__     = _f2_getattr_
    model.sp_integrate    = sp_integrate_2D
    model.sp_integrate_2D = sp_integrate_2D

Analysis = cpp.Analysis
for pdf in ( Analysis.Models.BreitWigner        , 
             Analysis.Models.Flatte             ,
             Analysis.Models.Bukin              ,
             Analysis.Models.PhaseSpace2        ,
             Analysis.Models.PhaseSpaceNL       ,
             Analysis.Models.PhaseSpace23L      ,
             Analysis.Models.PhaseSpaceLeft     ,
             Analysis.Models.PhaseSpaceRight    ,
             Analysis.Models.PhaseSpacePol      ,
             Analysis.Models.Needham            ,
             Analysis.Models.CrystalBall        ,
             Analysis.Models.CrystalBallRS      ,
             Analysis.Models.CrystalBallDS      , 
             Analysis.Models.Apolonios          ,
             Analysis.Models.Apolonios2         , 
             Analysis.Models.GramCharlierA      , 
             Analysis.Models.Voigt              ,
             Analysis.Models.PseudoVoigt        ,
             Analysis.Models.Logistic           ,
             Analysis.Models.LASS               ,
             Analysis.Models.Bugg               ,
             Analysis.Models.LASS23L            ,
             Analysis.Models.Bugg23L            , 
             Analysis.Models.BW23L              , 
             Analysis.Models.PolyPositive       ,
             Analysis.Models.ExpoPositive       ,
             Analysis.Models.TwoExpoPositive    ,
             Analysis.Models.PositiveSpline     ,
             Analysis.Models.MonothonicSpline   , 
             
             Analysis.Models.StudentT           ,
             Analysis.Models.BifurcatedStudentT , 
             Analysis.Models.GammaDist          , 
             Analysis.Models.GenGammaDist       , 
             Analysis.Models.Amoroso            ,
             Analysis.Models.LogGammaDist       ,
             Analysis.Models.Log10GammaDist     ,
             Analysis.Models.LogGamma           ,
             Analysis.Models.BetaPrime          ,
             Analysis.Models.Landau             ,
             Analysis.Models.SinhAsinh          , 
             Analysis.Models.JohnsonSU          ,
             Analysis.Models.Atlas              ,
             Analysis.Models.Sech               ,
             Analysis.Models.Swanson            ,
             Analysis.Models.Argus              ,
             Analysis.Models.Tsallis            ,
             Analysis.Models.QGSM               ,
             Analysis.Models.BifurcatedGauss    ,
             Analysis.Models.GenGaussV1         , 
             Analysis.Models.GenGaussV2         , 
             Analysis.Models.SkewGauss          ) :

    pdf.sp_integrate = sp_integrate_1D_

for pdf in ( Analysis.Models.Poly2DPositive     ,
             Analysis.Models.Poly2DSymPositive  , 
             Analysis.Models.PS2DPol            ,
             Analysis.Models.PS2DPolSym         , 
             Analysis.Models.ExpoPS2DPol        , 
             Analysis.Models.Expo2DPol          ,
             Analysis.Models.Expo2DPolSym       , 
             Analysis.Models.Spline2D           ,
             Analysis.Models.Spline2DSym        ) :
    
    pdf.sp_integrate = sp_integrate_2D_


#
## Decorate 3D models
# 
for model in ( Gaudi.Math.Bernstein3D    ,
               Gaudi.Math.Bernstein3DSym ,
               Gaudi.Math.Bernstein3DMix ,
               Gaudi.Math.Positive3D     ,
               Gaudi.Math.Positive3DSym  ,
               Gaudi.Math.Positive3DMix  ) :
    
    model . tf2 = _tf3_ 
    model . tf  = _tf3_ 
    model.sp_integrate = sp_integrate_3D
    model.__getattr__  = _f3_getattr_

for pdf in  ( Analysis.Models.Poly3DPositive    ,
              Analysis.Models.Poly3DSymPositive ,
              Analysis.Models.Poly3DMixPositive ) :
    
    pdf.sp_integrate = sp_integrate_3D_
              
# =============================================================================
## set parameter for polynomial/spline functions
#  @code
#  fun = ...
#  fun[1] = 10.0
#  @endcode 
def _p_set_par_ ( o , index , value ) :
    """Set parameter for polynomial/spline function
    >>> fun = ...
    >>> fun[1] = 10.0
    """
    return o.setPar ( index , value )

# =============================================================================
## get parameter from polynomial/spline functions
#  @code
#  fun = ...
#  print fun[1]
#  @endcode 
def _p_get_par_ ( o , index ) :
    """Get parameter from polynomial/spline function
    >>> fun = ...
    >>> print fun[1]
    """
    return o.par ( index )

# =============================================================================
## iterator over parameters of polynomial function
#  @code
#  fun = ...
#  for i in fun  : print fun[i] 
#  @endcode
def _p_iter_ ( o ) :
    """Iterator over parameters of polynomial function
    >>> fun = ...
    >>> for i in fun  : print fun[i] 
    """
    np = o.npars()
    for i in  range(np) :
        yield i
        
        
for f in ( Gaudi.Math.Positive        ,
           Gaudi.Math.PositiveEven    ,  
           Gaudi.Math.Bernstein       , 
           Gaudi.Math.BernsteinEven   , 
           Gaudi.Math.Bernstein2D     ,
           Gaudi.Math.Positive2D      ,
           Gaudi.Math.Bernstein2DSym  ,
           Gaudi.Math.Positive2DSym   ,
           ##
           Gaudi.Math.BSpline         ,
           Gaudi.Math.PositiveSpline  ,
           Gaudi.Math.Spline2D        ,
           Gaudi.Math.Spline2DSym     ,
           ##
           Gaudi.Math.Bernstein3D     ,
           Gaudi.Math.Bernstein3DSym  ,
           Gaudi.Math.Bernstein3DMix  ,
           Gaudi.Math.Positive3D      ,
           Gaudi.Math.Positive3DSym   ,
           Gaudi.Math.Positive3DMix   ,
           ## 
           Gaudi.Math.PolySum         ,
           ## 
           Gaudi.Math.NSphere         ) :

    f.__setitem__  = _p_set_par_
    f.__getitem__  = _p_get_par_
    f.__len__      = lambda s     : s.npars() 
    f.__iter__     = _p_iter_
    f.__contains__ = lambda s , i : 0<=i<len(s)

# =============================================================================
##  Long polynomial division
#   f(x) = q(x)*g(x)+r(x), where  deg(f)=m >= def(g)=n, and
#   deg(q)<=m-n, deg(r)<n
#   @code
#   f = ... # the first polynom
#   g = ... # the second polynom
#   q,r = divmod ( f , g ) 
#   @endcode
def _b_divmod_ ( b , p ) :
    """Long polynomial division
    f(x) = q(x)*g(x)+r(x), where  deg(f)=m >= def(g)=n, and
    deg(q)<=m-n, deg(r)<n
    
    >>> f = ... # the first polynom
    >>> g = ... # the second polynom
    >>> q,r = divmod ( f , g ) 
    """
    rr = b.divmod(p)
    return rr.first, rr.second

# =============================================================================
##  Long polynomial division
#   f(x) = q(x)*g(x)+r(x), where  deg(f)=m >= def(g)=n, and
#   deg(q)<=m-n, deg(r)<n
#   @code
#   f = ... # the first polynom
#   g = ... # thr second  polynom
#   q = f // g # get quotient 
#   @endcode
def _b_floordiv_ ( b , p ) :
    """Long polynomial division
    f(x) = q(x)*g(x)+r(x), where  deg(f)=m >= def(g)=n, and
    deg(q)<=m-n, deg(r)<n
    
    >>> f = ... # the first polynom
    >>> g = ... # the second polynom
    >>> q = f // g 
    """
    rr = b.divmod(p)
    return rr.first

# =============================================================================
##  Long polynomial division
#   f(x) = q(x)*g(x)+r(x), where  deg(f)=m >= def(g)=n, and
#   deg(q)<=m-n, deg(r)<n
#   @code
#   f = ... # the first polynom
#   g = ... # the second  polynom
#   r = f % g # get reminder 
#   @endcode
def _b_mod_ ( b , p ) :
    """Long polynomial division
    f(x) = q(x)*g(x)+r(x), where  deg(f)=m >= def(g)=n, and
    deg(q)<=m-n, deg(r)<n
    
    >>> f = ... # the first polynom
    >>> g = ... # the second polynom
    >>> r = f % g 
    """
    rr = b.divmod(p)
    return rr.second

# =============================================================================
## power function for polynomials
#  @code
#  fun1 = ..
#  fun2 = fun1**3  
#  @endcode
def _b_pow_ ( b , n ) :
    """Power function for polynomials
    >>> fun1 = ..
    >>> fun2 = fun1**3  
    """
    if n != int(n) : raise ValueError('Illegal non-integer  exponent:%s' % n )
    n = int(n) 
    if 0  > n      : raise ValueError('Illegal negative     exponent:%s' % n )
    return b.pow(n)

Gaudi.Math.Bernstein. __divmod__   = _b_divmod_
Gaudi.Math.Bernstein. __floordiv__ = _b_floordiv_
Gaudi.Math.Bernstein. __mod__      = _b_mod_
Gaudi.Math.Bernstein. __pow__      = _b_pow_
# ============================================================================

    
# =============================================================================
## random generators 
# =============================================================================
from random import uniform as _uniform_

# =============================================================================
## generate random numbers from bernstein-like distribuitions
#  @code
#  >>> func = ...
#  >>> for x in func.generate( 1000 ) : print x 
#  @endcode
def _random_generate_bernstein_ ( fun , num ) :
    """Generate random numbers from bernstein-like distribuitions
    >>> func = ...
    >>> for x in func.generate( 1000 ) : print x 
    """
    xmn = fun.xmin ()
    xmx = fun.xmax ()
    ymx = max ( fun.bernstein().pars() )
    i   = 0 
    while i < num : 
        x = _uniform_ ( xmn , xmx ) 
        y = _uniform_ (   0 , ymx )
        v = fun ( x )
        if v >= y :
            i+= 1 
            yield x
            
# =============================================================================
## Get random number from bernstein-like distribuitions
#  @code
#  >>> func = ...
#  >>> print fun.shoot() 
#  @endcode
def _random_shoot_bernstein_ ( fun ) :
    """Get random number from bernstein-like distribuitions
    >>> func = ...
    >>> print func.shoot()  
    """
    xmn = fun.xmin ()
    xmx = fun.xmax ()
    ymx = max ( fun.bernstein().pars() )
    i   = 0 
    while True : 
        x = _uniform_ ( xmn , xmx ) 
        y = _uniform_ (   0 , ymx )
        v = fun ( x )
        if v >= y : return x 



# =============================================================================
## generate random numbers from b-spline-distribuitions
#  @code
#  >>> func = ...
#  >>> for x in func.generate( 1000 ) : print x 
#  @endcode
def _random_generate_bspline_ ( fun , num ) :
    """Generate random numbers from bspline-like distribuitions
    >>> func = ...
    >>> for x in func.generate( 1000 ) : print x 
    """
    xmn = fun.xmin ()
    xmx = fun.xmax ()
    ymx = max ( fun.bspline().pars() )
    i   = 0 
    while i < num : 
        x = _uniform_ ( xmn , xmx ) 
        y = _uniform_ (   0 , ymx )
        v = fun ( x )
        if v >= y :
            i+= 1 
            yield x
            
# =============================================================================
## Get random number from bspline-like distribuitions
#  @code
#  >>> func = ...
#  >>> print fun.shoot() 
#  @endcode
def _random_shoot_bspline_ ( fun ) :
    """Get random number from bspline-like distribuitions
    >>> func = ...
    >>> print func.shoot()  
    """
    xmn = fun.xmin ()
    xmx = fun.xmax ()
    ymx = max ( fun.bspline().pars() )
    i   = 0 
    while True : 
        x = _uniform_ ( xmn , xmx ) 
        y = _uniform_ (   0 , ymx )
        v = fun ( x )
        if v >= y : return x 

# =============================================================================
## generate random numbers from 2D bernstein-like distribuitions
#  @code
#  >>> func = ...
#  >>> for x,y in func.generate( 1000 ) : print x,y 
#  @endcode
def _random_generate_bernstein2D_ ( fun , num = 1 ) :
    """Generate random numbers from 2D bernstein-like distribuitions
    >>> func = ...
    >>> for x,y in func.generate( 1000 ) : print x,y 
    """
    xmn = fun.xmin ()
    xmx = fun.xmax ()
    ymn = fun.ymin ()
    ymx = fun.ymax ()
    zmx = max ( fun.bernstein().pars() )
    i   = 0 
    while i < num : 
        x = random.uniform ( xmn , xmx ) 
        y = random.uniform ( ymn , ymx ) 
        x = random.uniform (   0 , zmx )
        v = fun ( x , y )
        if v >= z :
            i+= 1 
            yield x,y

# =============================================================================
## Get random number from 2D bernstein-like distribuitions
#  @code
#  >>> func = ...
#  >>> print fun.shoot() 
#  @endcode
def _random_shoot_bernstein2D_ ( fun ) :
    """Get random number from 2D bernstein-like distribuitions
    >>> func = ...
    >>> print func.shoot()  
    """
    xmn = fun.xmin ()
    xmx = fun.xmax ()
    ymn = fun.ymin ()
    ymx = fun.ymax ()
    zmx = max ( fun.bernstein().pars() )
    i   = 0 
    while True : 
        x = _uniform_ ( xmn , xmx ) 
        y = _uniform_ ( ymn , ymx ) 
        z = _uniform_ (   0 , zmx )
        v = fun ( x , y )
        if v >= z : return x,y

Gaudi.Math.Bernstein     .generate = _random_generate_bernstein_
Gaudi.Math.Bernstein     .shoot    = _random_shoot_bernstein_
Gaudi.Math.BernsteinEven .generate = _random_generate_bernstein_
Gaudi.Math.BernsteinEven .shoot    = _random_shoot_bernstein_


Gaudi.Math.Positive      .generate = _random_generate_bernstein_
Gaudi.Math.Positive      .shoot    = _random_shoot_bernstein_

Gaudi.Math.PositiveEven  .generate = _random_generate_bernstein_
Gaudi.Math.PositiveEven  .shoot    = _random_shoot_bernstein_

Gaudi.Math.PositiveSpline.generate = _random_generate_bspline_
Gaudi.Math.PositiveSpline.shoot    = _random_shoot_bspline_

Gaudi.Math.Positive2D    .generate = _random_generate_bernstein2D_
Gaudi.Math.Positive2D    .shoot    = _random_shoot_bernstein2D_
Gaudi.Math.Positive2DSym .generate = _random_generate_bernstein2D_
Gaudi.Math.Positive2DSym .shoot    = _random_shoot_bernstein2D_

# =============================================================================
## add complex amplitudes 
# =============================================================================
Gaudi.Math.LASS        . amp = _amp_
Gaudi.Math.LASS23L     . amp = _amp_
Gaudi.Math.Bugg23L     . amp = _amp_
Gaudi.Math.Flatte      . amp = _amp_
Gaudi.Math.Flatte2     . amp = _amp_
Gaudi.Math.Flatte23L   . amp = _amp_
Gaudi.Math.BreitWigner . amp = _amp_
    
# =============================================================================


import LHCbMath.deriv as _M
for i in ( _M.Derivative , _M.Integral , _M.IntegralCache ) :
    if not hasattr ( i , 'tf1' ) : i.tf1 = _tf1_
                
# =============================================================================
if '__main__' == __name__ :
    
    import Ostap.Line
    logger.info ( __file__  + '\n' + Ostap.Line.line  ) 
    logger.info ( 80*'*'   )
    logger.info ( __doc__  )
    logger.info ( 80*'*' )
    logger.info ( ' Author  : %s' %         __author__    ) 
    logger.info ( ' Date    : %s' %         __date__      )
    logger.info ( ' Symbols : %s' %  list ( __all__     ) )
    logger.info ( 80*'*' ) 
    
# =============================================================================
# The END 
# =============================================================================
