#!/usr/bin/env python
# -*- coding: utf-8 -*-
###############################################################################
# (c) Copyright 2000-2019 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
# =============================================================================
## @file TestReweigt.py
#
#  Test for reweighting machinery
# 
#  @author Vanya BELYAEV Ivan.Belyaev@itep.ru
#  @date 2014-05-11
# 
# =============================================================================
"""
Test for reweighting machinery
"""
# =============================================================================
__author__  = "Vanya BELYAEV Ivan.Belyaev@itep.ru"
__date__    = "2014-05-10"
__all__     = ()  ## nothing to be imported 
# =============================================================================
import ROOT, random, math, os, time 
from   Ostap.PyRoUts   import *
import Ostap.ZipShelve as     DBASE
import Ostap.TFileDeco
import Ostap.TreeDeco 
# =============================================================================
# logging 
# =============================================================================
from AnalysisPython.Logger import getLogger
if '__main__' == __name__  or '__builtin__'  == __name__ : 
    logger = getLogger ( 'Ostap.TestReweight' )
else : 
    logger = getLogger ( __name__ )
# =============================================================================    
logger.info ( 'Test for Reweighting machinery')
# =============================================================================
import  tempfile
testdata   =  os.path.join ( tempfile.gettempdir() , 'test_reweight_data.root' )
tag_data   = 'DATA_histogram'
tag_mc     = 'MC_tree'
dbname     = os.path.join ( tempfile.gettempdir() , 'test_reweight.db' )

## 
if os.path.exists ( testdata ) : os.remove ( testdata ) 
if os.path.exists ( dbname   ) : os.remove ( dbname   )

if not os.path.exists( testdata ) :

    seed = 1234567890L
    logger.info ( 'Test RANDOM data will be generated/seed=%s' % seed  )
    random.seed ( seed )
    
    ## prepare"data" histogram: lets use simple exponential and non-equidistance bins 
    hdata = h1_axis( [    i   for i in range(0,20) ] +
                     [ 20+i*2 for i in range(0,20) ] +
                     [ 60+i*4 for i in range(0,10) ] + [ 100 ] )
    
    for i in range( 0, 500000 ) :
        v = 100 + random.expovariate ( -1.0/60 ) 
        while v <   0 : v +=100
        while v > 100 : v -=100        
        hdata.Fill(v)        
    for i in range( 0, 500000 ) :
        v = random.gauss(50,10) 
        while v <   0 : v +=100
        while v > 100 : v -=100        
        hdata.Fill(v)
        
    with ROOT.TFile.Open( testdata ,'recreate') as mc_file:
        mc_file.cd() 
        ## write the histogram 
        mc_file[tag_data] = hdata

        mctree  = ROOT.TTree ( tag_mc , 'mc-tree' )
        mctree .SetDirectory ( mc_file ) 

        from array import array 
        xvar = array  ( 'f', [0])
        mctree.Branch ( 'x' , xvar , 'x/F' )
        
        for i in xrange ( 200000 ) : 
            xvar[0] = random.expovariate(1.0/80)            
            mctree.Fill()
            
        mctree.Write()

## Read data from 
dbroot = ROOT.TFile.open ( testdata , 'r' ) 
logger.info ( 'Test data is fetched from "%s"' % testdata )   
dbroot.ls()
hdata  = dbroot[ tag_data ]
mctree = dbroot[ tag_mc   ]

## prepare template histogram for MC 
hmc = ROOT.TH1D('hMC','histo-template for MC', 77 ,0,100 ) ; hmc.Sumw2() 

## prepare re-weighting machinery 
maxIter = 10  

## check database 
import os
if not os.path.exists( dbname ) :
    logger.info('Create new weights DBASE') 
    db = DBASE.open ( dbname , 'c' ) ##  create new empty db 
    db.close()
else :
    logger.info('Existing weights DBASE will be used') 

#
## make reweighting iterations
# 
from Ostap.Reweighting import Weight, makeWeights, WeightingPlot
from Ostap.Selectors   import SelectorWithVars, Variable  

## start iterations:
for iter in range ( 0 , maxIter ) :

    weighting = (
        ## variable          address in DB    
        Weight.Var( accessor = lambda s : s.x , address = 'x-reweight'  ) , 
        )
    
    weighter   = Weight( dbname , weighting )
    ## variables to be used in MC-dataset 
    variables  = [
        Variable ( 'pt_x'   , 'pt_x'   , 0  , 100 , lambda s : s.x ) , 
        Variable ( 'weight' , 'weight' , accessor = weighter       )  
        ]
    
    #
    ## create new "weighted" mcdataset
    # 
    selector = SelectorWithVars (
        variables ,
        '0<x && x<100 '
        )
    
    mctree.process ( selector )
    mcds = selector.data             ## new reweighted dataset

    #
    ## update weights
    #    
    plots    = [
        WeightingPlot( 'pt_x'   , 'weight' , 'x-reweight'  , hdata , hmc )  
        ]
    
    more = makeWeights ( mcds , plots , dbname , delta = 0.001 )

    ## make MC-histogram 
    mcds .project  ( hmc , 'pt_x' , 'weight'  )
    
    logger.info    ( 'Compare DATA and MC for iteration #%d' % iter )
    #
    ## compare the basic properties: mean, rms, skewness and kurtosis
    # 
    hdata.cmp_prnt ( hmc , 'DATA' , 'MC' , 'DATA vs MC' )
    #
    ## calculate the distances
    #
    dist = hdata.cmp_dist ( hmc , density = True )
    logger.info ('DATA-MC "distance"      %s' % dist )
    #
    ## calculate the 'orthogonality'
    #  
    cost = hdata.cmp_cos  ( hmc , density = True )
    logger.info ('DATA-MC "orthogonality" %s' % cost )
    #
    ## try to fit it DATA with MC and vice versa 
    #
    fit1 = hdata.cmp_fit ( hmc   , density = True )
    if fit1 and 0 == fit1.Status() :
        logger.info ( 'Fit DATA with MC   Prob=%.3g[%%] ' % ( fit1.Prob() * 100 ) )
    fit2 = hmc  .cmp_fit ( hdata , density = True )
    if fit2 and 0 == fit2.Status() :
        logger.info ( 'Fit MC   with DATA Prob=%.3g[%%] ' % ( fit2.Prob() * 100 ) )
    #
    ## make chi2-comparison between data and MC
    #
    c2ndf,prob = hdata.cmp_chi2 ( hmc   , density = True )
    logger.info ( 'DATA/MC: chi2/ndf (%.4g) and Prob %.5g%% ' % ( c2ndf , prob*100 ) )
    c2ndf,prob = hmc  .cmp_chi2 ( hdata , density  = True )
    logger.info ( 'MC/DATA: chi2/ndf (%.4g) and Prob %.5g%% ' % ( c2ndf , prob*100 ) )
    
    mn,mx = hdata.cmp_minmax ( hmc   , diff = lambda a,b : a/b , density = True )
    logger.info ("DATA*/MC   ``min/max-distance''[%%] (%s)/(%s) at x=%.1f/%.1f" % (
        (100*mn[1]-100).toString ( '%+.1f+-%.1f' ) ,
        (100*mx[1]-100).toString ( '%+.1f+-%.1f' ) , mn[0]  , mx[0] ) )
    mn,mx = hmc  .cmp_minmax ( hdata , diff = lambda a,b : b/a , density = True )
    logger.info ("DATA/MC*   ``min/max-distance''[%%] (%s)/(%s) at x=%.1f/%.1f" % (
        (100*mn[1]-100).toString ( '%+.1f+-%.1f' ) ,
        (100*mx[1]-100).toString ( '%+.1f+-%.1f' ) , mn[0]  , mx[0] ) )                 
     
    ## final density on data 
    data_density = hdata.density()
    ## final density on mc 
    mc_density   = hmc.density()
    data_density.red  ()
    mc_density  .blue ()
    data_density.draw ('e1')
    mc_density  .draw ('e1 same')
    time.sleep ( 5 ) 
    
    if not more : 
        logger.info    ( 'No more iterations are needed #%d' % iter )
        break
    
    if iter + 1 != maxIter :
        mcds.clear() 
        del mcds , selector
    else :
        del selector 


data_density.draw ('e1')
mc_density  .draw ('e1 same')
time.sleep(60)
dbroot.Close()


# =============================================================================
# The END 
# =============================================================================
