/*****************************************************************************\
* (c) Copyright 2000-2019 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
// ============================================================================
#ifndef ANALYSIS_STATVAR_H
#define ANALYSIS_STATVAR_H 1
// ============================================================================
// Include files
// ============================================================================
// STD & STL
// ============================================================================
#include <limits>
// ============================================================================
// GaudiKernel
// ============================================================================
#include "GaudiKernel/Kernel.h"
#include "GaudiKernel/StatEntity.h"
#include "GaudiKernel/SymmetricMatrixTypes.h"
// ============================================================================
// LHcbMath 
// ============================================================================
#include "LHCbMath/WStatEntity.h"
#include "LHCbMath/ValueWithError.h"
// ============================================================================
// Forward declarations 
// =============================================================================
class TTree  ;     // ROOT 
class TChain ;     // ROOT 
class TCut   ;     // ROOT 
// =============================================================================
class RooAbsData ; // RooFit 
// =============================================================================
namespace Analysis 
{
  // ==========================================================================
  /** @class StatVar Analysis/StatVar.h
   *  Helper class to get statistical 
   *  infomation  about the variable/expression 
   *
   *  @author Vanya BELYAEV Ivan.Belyaev@itep.ru
   *  @date   2013-10-13
   * 
   */
  class GAUDI_API StatVar 
  {
  public:
    // ========================================================================
    /// the actual type for statistic 
    typedef Gaudi::Math::WStatEntity  Statistic ;
    /// the actual type for interval 
    typedef std::pair<double,double>  Interval  ;
    // ========================================================================
  public: 
    // ========================================================================
    static_assert (std::numeric_limits<unsigned long>::is_specialized   ,
                   "Numeric_limist<unsigned long> are not specialized!" ) ;
    static constexpr unsigned long LAST { std::numeric_limits<unsigned long>::max() } ;
    // ========================================================================
  public: 
    // ========================================================================
    /** build statistic for the <code>expression</code>
     *  @param tree (INPUT) the tree 
     *  @param expression (INPUT) the expression
     *
     *  @code
     *  tree = ... 
     *  stat = tree.statVar ( 'S_sw' ) 
     *  @endcode 
     *
     *  @author Vanya BELYAEV Ivan.Belyaev@itep.ru
     *  @date   2013-10-13
     */
    static Statistic statVar
      ( TTree*              tree              , 
        const std::string&  expression        ,
        const unsigned long first      = 0    ,
        const unsigned long last       = LAST ) ;
    // ========================================================================
    /** build statistic for the <code>expression</code>
     *  @param tree       (INPUT) the tree 
     *  @param expression (INPUT) the expression
     *  @param cuts       (INPUT) the selection criteria 
     *
     *  @code
     *  tree = ... 
     *  stat = tree.statVar( 'S_sw' ,'pt>1000') 
     *  @endcode 
     *
     *  @author Vanya BELYAEV Ivan.Belyaev@itep.ru
     *  @date   2013-10-13
     */
    static Statistic statVar
      ( TTree*              tree              , 
        const std::string&  expression        , 
        const std::string&  cuts              ,
        const unsigned long first      = 0    ,
        const unsigned long last       = LAST ) ;
    // ========================================================================
    /** build statistic for the <code>expression</code>
     *  @param tree       (INPUT) the tree 
     *  @param expression (INPUT) the expression
     *  @param cuts       (INPUT) the selection criteria 
     *
     *  @code
     *  tree = ... 
     *  stat = tree.statVar( 'S_sw' ,'pt>1000') 
     *  @endcode 
     *
     *  @author Vanya BELYAEV Ivan.Belyaev@itep.ru
     *  @date   2013-10-13
     */
    static Statistic statVar
      ( TTree*              tree              , 
        const std::string&  expression        , 
        const TCut&         cuts              ,
        const unsigned long first      = 0    ,
        const unsigned long entries    = LAST ) ;
    // ========================================================================
    /** calculate the covariance of two expressions 
     *  @param tree  (INPUT)  the inpout tree 
     *  @param exp1  (INPUT)  the first  expresiion
     *  @param exp2  (INPUT)  the second expresiion
     *  @param stat1 (UPDATE) the statistic for the first  expression
     *  @param stat2 (UPDATE) the statistic for the second expression
     *  @param cov2  (UPDATE) the covariance matrix 
     *  @return number of processed events 
     *  
     *  @author Vanya BELYAEV Ivan.Belyaev@itep.ru
     *  @date   2014-03-27
     */
    static unsigned long statCov 
      ( TTree*               tree          , 
        const std::string&   exp1          , 
        const std::string&   exp2          , 
        Statistic&           stat1         ,  
        Statistic&           stat2         ,  
        Gaudi::SymMatrix2x2& cov2          , 
        const unsigned long  first  = 0    ,
        const unsigned long  last   = LAST ) ;
    // ========================================================================
    /** calculate the covariance of two expressions 
     *  @param tree  (INPUT)  the inpout tree 
     *  @param exp1  (INPUT)  the first  expresiion
     *  @param exp2  (INPUT)  the second expresiion
     *  @param cuts  (INPUT)  the selection criteria 
     *  @param stat1 (UPDATE) the statistic for the first  expression
     *  @param stat2 (UPDATE) the statistic for the second expression
     *  @param cov2  (UPDATE) the covariance matrix 
     *  @return number of processed events 
     *  
     *  @author Vanya BELYAEV Ivan.Belyaev@itep.ru
     *  @date   2014-03-27
     */
    static unsigned long statCov 
      ( TTree*               tree         ,
        const std::string&   exp1         , 
        const std::string&   exp2         , 
        const std::string&   cuts         ,
        Statistic&           stat1        ,   
        Statistic&           stat2        ,  
        Gaudi::SymMatrix2x2& cov2         , 
        const unsigned long  first = 0    ,
        const unsigned long  last  = LAST ) ;
    // ========================================================================
    /** calculate the covariance of two expressions 
     *  @param tree  (INPUT)  the inpout tree 
     *  @param exp1  (INPUT)  the first  expresiion
     *  @param exp2  (INPUT)  the second expresiion
     *  @param cuts  (INPUT)  the selection criteria 
     *  @param stat1 (UPDATE) the statistic for the first  expression
     *  @param stat2 (UPDATE) the statistic for the second expression
     *  @param cov2  (UPDATE) the covariance matrix 
     *  @return number of processed events 
     *  
     *  @author Vanya BELYAEV Ivan.Belyaev@itep.ru
     *  @date   2014-03-27
     */
    static unsigned long statCov 
      ( TTree*               tree         ,
        const std::string&   exp1         , 
        const std::string&   exp2         , 
        const TCut&          cuts         ,
        Statistic&           stat1        ,  
        Statistic&           stat2        ,  
        Gaudi::SymMatrix2x2& cov2         , 
        const unsigned long  first = 0    ,
        const unsigned long  last  = LAST ) ;
    // ========================================================================
  public:
    // ========================================================================
    /** build statistic for the <code>expression</code>
     *  @param data       (INPUT) the data 
     *  @param expression (INPUT) the expression
     *  @param first      (INPUT) the frist event to process
     *  @param last       (INPUT) process till "last"-event
     *
     *  @code
     *  data = ... 
     *  stat = data.statVar( 'S_sw' , 'pt>10') 
     *  @endcode 
     *
     *  @author Vanya BELYAEV Ivan.Belyaev@itep.ru
     *  @date   2015-02-15
     */
    static Statistic statVar 
      ( const RooAbsData*   data               , 
        const std::string&  expression         , 
        const unsigned long first      = 0     ,
        const unsigned long last       = LAST  ) 
    { return statVar ( data , expression , std::string() , std::string() , first , last ) ; }
    // ========================================================================
    /** build statistic for the <code>expression</code>
     *  @param data       (INPUT) the data 
     *  @param expression (INPUT) the expression
     *  @param cuts       (INPUT) the selection
     *  @param first      (INPUT) the frist event to process
     *  @param last       (INPUT) process till "last"-event
     *
     *  @code
     *  data = ... 
     *  stat = data.statVar( 'S_sw' , 'pt>10') 
     *  @endcode 
     *
     *  @author Vanya BELYAEV Ivan.Belyaev@itep.ru
     *  @date   2015-02-15
     */
    static Statistic statVar 
      ( const RooAbsData*   data               , 
        const std::string&  expression         , 
        const std::string&  cuts               , 
        const unsigned long first      = 0     ,
        const unsigned long last       = LAST  ) 
    { return statVar ( data , expression , cuts , std::string() , first , last ) ; }
    // ========================================================================
    /** build statistic for the <code>expression</code>
     *  @param data       (INPUT) the data 
     *  @param expression (INPUT) the expression
     *  @param cuts       (INPUT) the selection
     *  @param first      (INPUT) the frist event to process
     *  @param last       (INPUT) process till "last"-event
     *
     *  @code
     *  data = ... 
     *  stat = data.statVar( 'S_sw' , 'pt>10') 
     *  @endcode 
     *
     *  @author Vanya BELYAEV Ivan.Belyaev@itep.ru
     *  @date   2015-02-15
     */
    static Statistic statVar 
      ( const RooAbsData*   data               , 
        const std::string&  expression         , 
        const TCut&         cuts               , 
        const unsigned long first      = 0     ,
        const unsigned long last       = LAST  ) 
    { return statVar ( data , expression , cuts , std::string() , first , last ) ; }
    // ========================================================================
    /** build statistic for the <code>expression</code>
     *  @param data       (INPUT) the data 
     *  @param expression (INPUT) the expression
     *  @param cuts       (INPUT) the selection
     *  @param cutrange   (INPUT) the cut-range
     *  @param first      (INPUT) the frist event to process
     *  @param last       (INPUT) process till "last"-event
     *
     *  @code
     *  data = ... 
     *  stat = data.statVar( 'S_sw' , 'pt>10') 
     *  @endcode 
     *
     *  @author Vanya BELYAEV Ivan.Belyaev@itep.ru
     *  @date   2015-02-15
     */
    static Statistic statVar 
      ( const RooAbsData*   data               , 
        const std::string&  expression         , 
        const TCut&         cuts               , 
        const std::string&  cut_range          ,
        const unsigned long first      = 0     ,
        const unsigned long last       = LAST  ) ;
    // ========================================================================
    /** build statistic for the <code>expression</code>
     *  @param data       (INPUT) the data 
     *  @param expression (INPUT) the expression
     *  @param cuts       (INPUT) the selection
     *  @param cutrange   (INPUT) the cut-range
     *  @param first      (INPUT) the frist event to process
     *  @param last       (INPUT) process till "last"-event
     *
     *  @code
     *  data = ... 
     *  stat = data.statVar( 'S_sw' , 'pt>10') 
     *  @endcode 
     *
     *  @author Vanya BELYAEV Ivan.Belyaev@itep.ru
     *  @date   2015-02-15
     */
    static Statistic statVar 
      ( const RooAbsData*   data               , 
        const std::string&  expression         , 
        const std::string&  cuts               , 
        const std::string&  cut_range          ,
        const unsigned long first      = 0     ,
        const unsigned long last       = LAST  ) ;
    // ========================================================================
  public:
    // ========================================================================
    /** calculate the covariance of two expressions 
     *  @param tree  (INPUT)  the input  tree 
     *  @param exp1  (INPUT)  the first  expresiion
     *  @param exp2  (INPUT)  the second expresiion
     *  @param stat1 (UPDATE) the statistic for the first  expression
     *  @param stat2 (UPDATE) the statistic for the second expression
     *  @param cov2  (UPDATE) the covariance matrix 
     *  @return number of processed events 
     *  
     *  @author Vanya BELYAEV Ivan.Belyaev@itep.ru
     *  @date   2014-03-27
     */
    static unsigned long statCov
      ( const RooAbsData*    tree             , 
        const std::string&   exp1             , 
        const std::string&   exp2             , 
        Statistic&           stat1            ,  
        Statistic&           stat2            ,  
        Gaudi::SymMatrix2x2& cov2             , 
        const std::string&   cut_range = ""   ,
        const unsigned long  first     = 0    ,
        const unsigned long  last      = LAST ) ;
    // ========================================================================
    /** calculate the covariance of two expressions 
     *  @param tree  (INPUT)  the input  tree 
     *  @param exp1  (INPUT)  the first  expresiion
     *  @param exp2  (INPUT)  the second expresiion
     *  @param cuts  (INPUT)  the selection 
     *  @param stat1 (UPDATE) the statistic for the first  expression
     *  @param stat2 (UPDATE) the statistic for the second expression
     *  @param cov2  (UPDATE) the covariance matrix 
     *  @return number of processed events 
     *  
     *  @author Vanya BELYAEV Ivan.Belyaev@itep.ru
     *  @date   2014-03-27
     */
    static unsigned long statCov
      ( const RooAbsData*    tree             , 
        const std::string&   exp1             , 
        const std::string&   exp2             , 
        const std::string&   cuts             , 
        Statistic&           stat1            ,  
        Statistic&           stat2            ,  
        Gaudi::SymMatrix2x2& cov2             , 
        const std::string&   cut_range = ""   ,
        const unsigned long  first     = 0    ,
        const unsigned long  last      = LAST ) ;
    // ========================================================================
  public:
    // ========================================================================
    /** get the number of equivalent entries 
     *  \f$ n_{eff} \equiv = \frac{ (\sum w)^2}{ \sum w^2} \f$
     *  @param tree  (INPUT) the tree 
     *  @param cuts  (INPUT) selection  criteria 
     *  @param  first  (INPUT) the first  event to process 
     *  @param  last   (INPUT) the last event to  process
     *  @return number of equivalent entries 
     */
    static double nEff 
      ( TTree&               tree          , 
        const std::string&   cuts   = ""   , 
        const unsigned long  first  = 0    ,
        const unsigned long  last   = LAST ) ;
    // ========================================================================
    /** get the number of equivalent entries 
     *  \f$ n_{eff} \equiv = \frac{ (\sum w)^2}{ \sum w^2} \f$
     *  @param tree      (INPUT) the tree 
     *  @param cuts      (INPUT) selection crietria  
     *  @param first     (INPUT) the first  event to process 
     *  @param last      (INPUT) the last event to  process
     *  @param cut_range (INPUT) cut range 
     *  @return number of equivalent entries 
     */
    static double nEff 
      ( const RooAbsData&    tree             , 
        const std::string&   cuts      = ""   , 
        const std::string&   cut_range = ""   , 
        const unsigned long  first     = 0    ,
        const unsigned long  last      = LAST ) ;
    // ========================================================================
  public:
    // ========================================================================
    /** calculate the moment of order "order" relative to the center "center"
     *  @param  tree   (INPUT) input tree 
     *  @param  expr   (INPUT) expression  (must  be valid TFormula!)
     *  @param  order  (INPUT) the order 
     *  @param  center (INPUT) the center 
     *  @param  cuts   (INPUT) cuts 
     *  @param  first  (INPUT) the first  event to process 
     *  @param  last   (INPUT) the last event to  process
     *  @return the moment 
     */
    static double get_moment 
      ( TTree&               tree          ,  
        const unsigned short order         , 
        const std::string&   expr          , 
        const double         center = 0    ,
        const std::string&   cuts   = ""   , 
        const unsigned long  first  = 0    ,
        const unsigned long  last   = LAST ) ;
    // ========================================================================    
    /** calculate the moment of order "order" relative to the center "center"
     *  @param  data      (INPUT) input data
     *  @param  expr      (INPUT) expression  (must  be valid TFormula!)
     *  @param  order     (INPUT) the order 
     *  @param  center    (INPUT) the center 
     *  @param  cuts      (INPUT) cuts 
     *  @param  first     (INPUT) the first  event to process 
     *  @param  last      (INPUT) the last event to  process
     *  @param  cut_range (INPUT) cut range 
     *  @return the moment 
     */
    static double get_moment 
      ( const RooAbsData&    data             ,  
        const unsigned short order            , 
        const std::string&   expr             , 
        const double         center    = 0    ,
        const std::string&   cuts      = ""   , 
        const std::string&   cut_range = ""   , 
        const unsigned long  first     = 0    ,
        const unsigned long  last      = LAST ) ;
    // ========================================================================    
  public:
    // ========================================================================    
    /** calculate the moment of order "order"
     *  @param  tree  (INPUT) input tree 
     *  @param  order (INPUT) the order 
     *  @param  expr  (INPUT) expression  (must  be valid TFormula!)
     *  @param  cuts  (INPUT) cuts 
     *  @param  first (INPUT) the first  event to process 
     *  @param  last  (INPUT) the last event to  process
     *  @return the moment 
     */ 
    static Gaudi::Math::ValueWithError moment
      ( TTree&               tree         ,  
        const unsigned short order        ,
        const std::string&   expr         , 
        const std::string&   cuts  = ""   , 
        const unsigned long  first = 0    ,
        const unsigned long  last  = LAST ) ;
    // ========================================================================
    /** calculate the moment of order "order"
     *  @param  data  (INPUT) input data
     *  @param  order (INPUT) the order 
     *  @param  expr  (INPUT) expression  (must  be valid TFormula!)
     *  @param  cuts  (INPUT) cuts 
     *  @param  first (INPUT) the first  event to process 
     *  @param  last  (INPUT) the last event to  process
     *  @param  cut_range (INPUT) cut range 
     *  @return the moment 
     */ 
    static Gaudi::Math::ValueWithError moment
      ( const  RooAbsData&   data             ,  
        const unsigned short order            ,
        const std::string&   expr             , 
        const std::string&   cuts      = ""   , 
        const std::string&   cut_range = ""   , 
        const unsigned long  first     = 0    ,
        const unsigned long  last      = LAST ) ;
    // ========================================================================
  public:
    // ========================================================================
    /** calculate the central moment of order "order"
     *  @param  tree  (INPUT) input tree 
     *  @param  order (INPUT) the order 
     *  @param  expr  (INPUT) expression  (must  be valid TFormula!)
     *  @param  cuts  (INPUT) cuts 
     *  @param  first (INPUT) the first  event to process 
     *  @param  last  (INPUT) the last event to  process
     *  @return the moment 
     */ 
    static Gaudi::Math::ValueWithError central_moment 
      ( TTree&               tree         ,  
        const unsigned short order        ,
        const std::string&   expr         , 
        const std::string&   cuts  = ""   , 
        const unsigned long  first = 0    ,
        const unsigned long  last  = LAST ) ;
    // ========================================================================
    /** calculate the central moment of order "order"
     *  @param  data  (INPUT) input data
     *  @param  order (INPUT) the order 
     *  @param  expr  (INPUT) expression  (must  be valid TFormula!)
     *  @param  cuts  (INPUT) cuts 
     *  @param  first (INPUT) the first  event to process 
     *  @param  last  (INPUT) the last event to  process
     *  @param  cut_range (INPUT) cut range 
     *  @return the moment 
     */ 
    static Gaudi::Math::ValueWithError central_moment 
      ( const RooAbsData&    data             ,  
        const unsigned short order            ,
        const std::string&   expr             , 
        const std::string&   cuts      = ""   , 
        const std::string&   cut_range = ""   , 
        const unsigned long  first     = 0    ,
        const unsigned long  last      = LAST ) ;
    // ========================================================================    
  public:
    // ========================================================================    
    /** calculate the skewness of the  distribution
     *  @param  tree  (INPUT) input tree 
     *  @param  expr  (INPUT) expression  (must  be valid TFormula!)
     *  @param  cuts  (INPUT) cuts 
     *  @param  first (INPUT) the first  event to process 
     *  @param  last  (INPUT) the last event to  process
     *  @return the skewness 
     */
    static Gaudi::Math::ValueWithError skewness 
      ( TTree&               tree         ,  
        const std::string&   expr         , 
        const std::string&   cuts  = ""   , 
        const unsigned long  first = 0    ,
        const unsigned long  last  = LAST ) ;
    // ========================================================================    
    /** calculate the skewness of the  distribution
     *  @param  data      (INPUT) input data
     *  @param  expr      (INPUT) expression  (must  be valid TFormula!)
     *  @param  cuts      (INPUT) cuts 
     *  @param  first     (INPUT) the first  event to process 
     *  @param  last      (INPUT) the last event to  process
     *  @param  cut_range (INPUT) cut range 
     *  @return the skewness 
     */
    static Gaudi::Math::ValueWithError skewness 
      ( const RooAbsData&    data             ,  
        const std::string&   expr             , 
        const std::string&   cuts      = ""   , 
        const std::string&   cut_range = ""   , 
        const unsigned long  first     = 0    ,
        const unsigned long  last      = LAST ) ;
    // ========================================================================    
  public:
    // ========================================================================    
    /** calculate the (excess) kurtosis of the  distribution
     *  @param  tree  (INPUT) input tree 
     *  @param  expr  (INPUT) expression  (must  be valid TFormula!)
     *  @param  cuts  (INPUT) cuts 
     *  @param  first (INPUT) the first  event to process 
     *  @param  last  (INPUT) the last event to  process
     *  @return the (excess) kurtosis
     */
    static Gaudi::Math::ValueWithError kurtosis
      ( TTree&               tree         ,  
        const std::string&   expr         , 
        const std::string&   cuts  = ""   , 
        const unsigned long  first = 0    ,
        const unsigned long  last  = LAST ) ;
    // ========================================================================    
    /** calculate the (excess) kurtosis of the  distribution
     *  @param  data  (INPUT) input data
     *  @param  expr  (INPUT) expression  (must  be valid TFormula!)
     *  @param  cuts  (INPUT) cuts 
     *  @param  first (INPUT) the first  event to process 
     *  @param  last  (INPUT) the last event to  process
     *  @param  cut_range (INPUT) cut range 
     *  @return the (excess) kurtosis
     */
    static Gaudi::Math::ValueWithError kurtosis
      ( const RooAbsData&    data             ,  
        const std::string&   expr             , 
        const std::string&   cuts      = ""   , 
        const std::string&   cut_range = ""   , 
        const unsigned long  first     = 0    ,
        const unsigned long  last      = LAST ) ;
    // ========================================================================    
  public:
    // ========================================================================    
    /**  get quantile of the distribution  
     *   @param tree  (INPUT) the input tree 
     *   @param q     (INPUT) quantile value   0 < q < 1  
     *   @param expr  (INPUT) the expression 
     *   @param cuts  (INPUT) selection cuts 
     *   @param  first (INPUT) the first  event to process 
     *   @param  last  (INPUT) the last event to  process
     *   @return the quantile value 
     */
    static double quantile
      ( TTree&              tree         ,
        const double        q            , //  0<q<1 
        const std::string&  expr         , 
        const std::string&  cuts  = ""   , 
        const unsigned long first = 0    ,
        const unsigned long last  = LAST ) ;
    // ========================================================================    
    /**  get quantile of the distribution  
     *   @param data   (INPUT) the input data
     *   @param q      (INPUT) quantile value   0 < q < 1  
     *   @param expr   (INPUT) the expression 
     *   @param cuts   (INPUT) selection cuts 
     *   @param  first (INPUT) the first  event to process 
     *   @param  last  (INPUT) the last event to  process
     *   @param  cut_range (INPUT) cut range 
     *   @return the quantile value 
     */
    static double quantile
      ( const RooAbsData&   data             ,
        const double        q                , //  0<q<1 
        const std::string&  expr             , 
        const std::string&  cuts      = ""   , 
        const std::string&  cut_range = ""   , 
        const unsigned long first     = 0    ,
        const unsigned long last      = LAST ) ;
    // ========================================================================    
  public:
    // ========================================================================    
    /**  get quantiles of the distribution  
     *   @param tree  (INPUT) the input tree 
     *   @param q     (INPUT) quantile value   0 < q < 1  
     *   @param expr  (INPUT) the expression 
     *   @param cuts  (INPUT) selection cuts 
     *   @param  first (INPUT) the first  event to process 
     *   @param  last  (INPUT) the last event to  process
     *   @return the quantile value 
     */
    static std::vector<double> quantiles
      ( TTree&                     tree             ,
        const std::vector<double>& quantiles        , 
        const std::string&         expr             , 
        const std::string&         cuts      = ""   , 
        const unsigned long        first     = 0    ,
        const unsigned long        last      = LAST ) ;
    // ========================================================================
    /**  get quantiles of the distribution  
     *   @param data  (INPUT) the input data
     *   @param q     (INPUT) quantile value   0 < q < 1  
     *   @param expr  (INPUT) the expression 
     *   @param cuts  (INPUT) selection cuts 
     *   @param  first (INPUT) the first  event to process 
     *   @param  last  (INPUT) the last event to  process
     *   @return the quantile value 
     */
    static std::vector<double> quantiles
      ( const RooAbsData&          data              ,
        const std::vector<double>& quantiles         , 
        const std::string&         expr              , 
        const std::string&         cuts       = ""   , 
        const std::string&         cut_range  = ""   , 
        const unsigned long        first      = 0    ,
        const unsigned long        last       = LAST ) ;
    // ========================================================================
  public:
    // ========================================================================    
    /**  get the interval of the distribution  
     *   @param tree  (INPUT) the input tree 
     *   @param q1    (INPUT) quantile value   0 < q1 < 1  
     *   @param q2    (INPUT) quantile value   0 < q2 < 1  
     *   @param expr  (INPUT) the expression 
     *   @param cuts  (INPUT) selection cuts 
     *   @param  first (INPUT) the first  event to process 
     *   @param  last  (INPUT) the last event to  process
     *   @return the quantile value 
     *   @code
     *   Tree& tree = ... ;
     *   /// get 90% interval:
     *   Interval ab = interval ( tree , 0.05 , 0.95 , 'mass' , 'pt>3' ) ;
     *   @code 
     */
    static Interval interval 
      ( TTree&              tree         ,
        const double        q1           , //  0<q1<1 
        const double        q2           , //  0<q2<1 
        const std::string&  expr         , 
        const std::string&  cuts  = ""   , 
        const unsigned long first = 0    ,
        const unsigned long last  = LAST ) ;
    // ========================================================================    
    /**  get the interval of the distribution  
     *   @param data  (INPUT) the input data
     *   @param q1    (INPUT) quantile value   0 < q1 < 1  
     *   @param q2    (INPUT) quantile value   0 < q2 < 1  
     *   @param expr  (INPUT) the expression 
     *   @param cuts  (INPUT) selection cuts 
     *   @param  first (INPUT) the first  event to process 
     *   @param  last  (INPUT) the last event to  process
     *   @return the quantile value 
     *   @code
     *   const RooAbsData& data = ... ;
     *   /// get 90% interval:
     *   Interval ab = interval ( data , 0.05 , 0.95 , 'mass' , 'pt>3' ) ;
     *   @code 
     */
    static Interval interval 
      ( const RooAbsData&   data             ,
        const double        q1               , //  0<q1<1 
        const double        q2               , //  0<q2<1 
        const std::string&  expr             , 
        const std::string&  cuts      = ""   , 
        const std::string&  cut_range = ""   , 
        const unsigned long first     = 0    ,
        const unsigned long last      = LAST ) ;
    // ========================================================================    
  };
  // ==========================================================================
} //                                                  end of namespace Analysis
// ============================================================================
//                                                                      The END 
// ============================================================================
#endif // ANALYSIS_STATVAR_H
// ============================================================================
