/*****************************************************************************\
* (c) Copyright 2000-2019 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
// ============================================================================
// Include
// ============================================================================
#include <memory>
#include <fstream>
#include <streambuf>
#include <iostream>
// ============================================================================
// Local
// ============================================================================
#include "Analysis/Tee.h"
// ============================================================================
/** @file  Tee.cpp
 *  Implementation file for class Gaudi::Utils::Tee
 *  @see Gaudi::Utils::Tee
 *  @see Gaudi::Utils::Mute
 *  @author Vanya BELYAEV Ivan.Belyaev@itep.ru
 *  @date 2013-07-07
 */
// ============================================================================
namespace std
{
  // ==========================================================================
  /** @class basic_teebuf
   *  implementation of "tee"-functionality
   *  the actualcode is copied from
   *  http://wordaligned.org/articles/cpp-streambufs
   */
  template <typename CHAR_TYPE, typename TRAITS = std::char_traits<CHAR_TYPE> >
  class basic_teebuf: public std::basic_streambuf<CHAR_TYPE,TRAITS>
  {
  public:
    // ========================================================================
    typedef typename TRAITS::int_type int_type;
    // ========================================================================
  public:
    // ========================================================================
    basic_teebuf
    ( std::basic_streambuf<CHAR_TYPE,TRAITS> * sb1 ,
      std::basic_streambuf<CHAR_TYPE,TRAITS> * sb2 )
      : sb1(sb1)
      , sb2(sb2)
    {}
    // ========================================================================
  private:
    // ========================================================================
    int sync() override
    {
      int const r1 = sb1->pubsync();
      int const r2 = sb2->pubsync();
      return r1 == 0 && r2 == 0 ? 0 : -1;
    }
    //
    int_type overflow ( int_type c ) override
    {
      //
      int_type const eof = TRAITS::eof();
      //
      if ( TRAITS::eq_int_type(c, eof)) { return TRAITS::not_eof(c); }
      else
      {
        //
        CHAR_TYPE const ch = TRAITS::to_char_type(c);
        int_type  const r1 = sb1->sputc(ch);
        int_type  const r2 = sb2->sputc(ch);
        return
          TRAITS::eq_int_type(r1, eof) ||
          TRAITS::eq_int_type(r2, eof) ? eof : c;
        //
      }
    }
    // ========================================================================
  private:
    // ========================================================================
    std::basic_streambuf<CHAR_TYPE,TRAITS> *sb1 ;
    std::basic_streambuf<CHAR_TYPE,TRAITS> *sb2 ;
    // ========================================================================
  };
  //
  // ==========================================================================
  /** @class basic_teestream
   *  implementation of "tee"-functionality
   *  the actualcode is copied from
   *  http://wordaligned.org/articles/cpp-streambufs
   */
  template < typename CHAR_TYPE, typename TRAITS = std::char_traits<CHAR_TYPE> >
  struct basic_teestream : public std::basic_ostream<CHAR_TYPE,TRAITS>
  {
    // ========================================================================
    typedef std::basic_ostream<CHAR_TYPE,TRAITS> stream_type ;
    typedef std::basic_teebuf <CHAR_TYPE,TRAITS> streambuff_type ;
    // ========================================================================
    basic_teestream ( stream_type& first, stream_type& second )
      : stream_type(&stmbuf), stmbuf( first.rdbuf(), second.rdbuf() ) {}
    //
    basic_teestream ( streambuff_type* first, streambuff_type* second )
      : stream_type(&stmbuf), stmbuf(first,second ) {}
    //
    ~basic_teestream() { stmbuf.pubsync() ; }
    // ========================================================================
  private:
    // ========================================================================
    streambuff_type stmbuf ;
    // ========================================================================
  };
  // ==========================================================================
  typedef basic_teebuf<char>    teebuf    ;
  typedef basic_teestream<char> teestream ;
  // ==========================================================================
}
// ============================================================================

// ============================================================================
// constructor from the filename
// ============================================================================
Gaudi::Utils::Tee::Tee  ( const std::string&  filename   )
  : m_file   ( new std::ofstream ( filename.c_str() ) )
  , m_own    ( true    )
  , m_buffer ( nullptr )
  , m_keep   ( nullptr )
{
  m_keep   = std::cout.rdbuf() ;
  m_buffer.reset ( new std::teebuf ( m_keep , m_file->rdbuf() ) ) ;
  std::cout.rdbuf ( m_buffer.get() ) ;
}
// ============================================================================
// constructor from the stream
// ============================================================================
Gaudi::Utils::Tee::Tee  ( std::ostream&  filestream )
  : m_file   ( &filestream )
  , m_own    ( false       )
  , m_buffer ( nullptr )
  , m_keep   ( nullptr )
{
  m_keep   = std::cout.rdbuf() ;
  m_buffer.reset ( new std::teebuf ( m_keep , m_file->rdbuf() ) ) ;
  std::cout.rdbuf ( m_buffer.get() ) ;
}
// ============================================================================
// destructor
// ============================================================================
Gaudi::Utils::Tee::~Tee() { exit () ; }
// ============================================================================
/*  helper function to implement python's __enter__
 *  the action is performed in constructor
 */
// ============================================================================
void Gaudi::Utils::Tee::enter (){}
// ============================================================================
//  helper function to implement python's __exit__
// ============================================================================
void Gaudi::Utils::Tee::exit  ()
{
  // ==========================================================================
  // avoid double exit
  if ( nullptr == m_keep ) { return ; }
  //
  // 1. restore std::cout
  std::cout.rdbuf ( m_keep ) ;
  // 2. delete the buffer
  m_buffer.reset() ;
  // 3. delete/close  the file (if needed)
  if   ( m_own ) { m_file.reset   () ; }
  else           { m_file.release () ; }
  //
  m_keep = 0 ;
  // ==========================================================================
}
// ============================================================================
//  The END
// ============================================================================


